--[[
Features:
LibQuickHealth-2.0 is a library that provides more up to date health data than the default Blizz events and functions.
In addition to the standard UNIT_HEALTH event, the library listens to combat log events (CLEU) to update health values. 
CLEU events fire more often than UNIT_HEALTH, and allow the library to get around the ~300ms throttling of UNIT_HEALTH updates.
The interesting CLEU events are those containing health differences (damage and heals). Normally, when UNIT_HEALTH fires,
those differences add up to the value reported by Blizz's UnitHealth. But sometimes, when a unit takes damage according to the CLEU,
that damage is processed *after* the next UNIT_HEALTH event.
This is called "pending damage". One of the features of LibQuickHealth-2.0 (and an improvement over 1.0) is to detect pending damage,
and incorporate it into the health values that are presented to players.

Limitations:
1.) Sometimes there are combat log entries like this: -6000 (damage), +5000 (heal).
But what really happened on the server was that the heal occured first.
If at least part of this heal was overheal, the prediction (-6000+5000 = -1000) is way off,
because what happened is for example +50 heal first(4950 overheal), and then 6000 damage,
for a total of +50-6000 = -5950. No AddOn could possibly prevent or predict this, but luckily these cases are very rare. 
However, if you see something like this happening (tank at ~100%, then takes damage instantly followed by a heal),
you should know that this heal might have been overheal. So don't cancel your heal. ;)

2.) If the maximum health changes, pending damage is completely ignored. Example: The tank is at 11001/18000,
takes 9001 damage and uses Last Stand at the same time.
Your client might tell you now that the tank health is 16401/23400 (+30% health and max health),
because that 9k hit was not yet processed (pending damage).
What the library does is: Show the tank at 11001/18000, then at 11001-9001 = 2000/18000, then at 16401/23400,
because pending damage is ignored. But on the server, the tank has only 16401-9001 = 7400/23400 health!
This looks very similar to the case before: The tank takes damage, but shortly afterwards that damage is (magically) undone.
Again, if you see this, don't cancel your heal!

3.) Bloodrush. There's no combat log entry for the health lost.
I believe the health cost depends on the caster level and race, but I don't know the exact formula.

4.) Exceptionally large HP5 values. There are no combat log entries for HP5 ticks, which occur every 2 seconds.
The library assumes that the health regen in combat is between 0 and 50 HP5, i.e. HP5 ticks between 0 and 20.
Common HP5 values: Demon skin/armor <=18 (24 talented?), Elixir of Major Fortitude 10, Enchant Boots - Vitality 4.
Less common: Dreaming Glory 30 (buff from herbalism), and a few level 60 items. 
So the library might get confused by two things: Small amounts of pending damage (20 or less), and Warlocks with herbalism...

Usage:
The library supports two events, UnitHealthUpdated and HealthUpdated.
The difference is that HealthUpdated fires for
guids, and UnitHealthUpdated fires for unitIDs.
UnitHealthUpdated replaces not only UNIT_HEALTH, but also UNIT_MAXHEALTH.
That means even if only the max health changes (but not the health),
UnitHealthUpdated (and HealthUpdated) will fire. QuickHealth.UnitHealth replaces Blizz's UnitHealth. See example 2 below.

To start listening to the event, do QuickHealth.RegisterCallback(YourAddonTableHere, "HealthUpdated", HandlerMethodHere).
The "HealthUpdated" callback triggers with 5 arguments: self, event, guid, health, healthMax (first two passed by the callbackhandler).
The "UnitHealthUpdated" callback triggers with 5 arguments: self, event, unitID, health, healthMax.
Additionaly, QuickHealth.UnitHealth(unitID) can be called to get what QuickHealth thinks is the current health.

Example 1: Using UnitHealthUpdated.
	local QuickHealth = LibStub("LibQuickHealth-2.0")
	local MyAddon = {}
	function MyAddon:Initialize()
		-- Note: It's QuickHealth.RegisterCallback(), not QuickHealth:RegisterCallback().
		QuickHealth.RegisterCallback(MyAddon, "UnitHealthUpdated", "UnitHealthUpdated")
	end
	function MyAddon:UnitHealthUpdated(event, unitID, health, healthMax)
		assert(event == "UnitHealthUpdated")
		ChatFrame1:AddMessage(format("%s HealthUpdated %i/%i", unitID, health, healthMax))
		assert(health==LibQuickHealth.UnitHealth(unitID))
	end


Example 2: Making LibQuickHealth an optional dependency.
	local QuickHealth = LibStub and LibStub("LibQuickHealth-2.0", true) -- don't error if not found
	local UnitHealth = QuickHealth and QuickHealth.UnitHealth or UnitHealth
	MyAddon = ...
	function MyAddon:Initialize()
		if QuickHealth then
			-- Add QuickHealth support. Register for the event and add a simple adapter method.
			QuickHealth.RegisterCallback(MyAddon, "UnitHealthUpdated", "UnitHealthUpdated")
			function MyAddon:UnitHealthUpdated(event, unitID)
				self:UpdateStuff(unitID)
			end
		else
			-- Assuming it's an Ace addon.
			MyAddon:RegisterEvent("UNIT_HEALTH", "UpdateStuff")
			MyAddon:RegisterEvent("UNIT_MAXHEALTH", "UpdateStuff")
		end
	end
	function MyAddon:UpdateStuff(unitID)
		local hp, max = UnitHealth(unitID), UnitHealthMax(unitID)
		-- do something with hp and max
	end


Example 3: Using HealthUpdated.
	local QuickHealth = LibStub("LibQuickHealth-2.0")
	local MyAddon = {}
	function MyAddon:Initialize()
		QuickHealth.RegisterCallback(MyAddon, "HealthUpdated", "HealthUpdated")
	end
	function MyAddon:HealthUpdated(event, guid, health, healthMax)
		assert(event == "HealthUpdated")
		ChatFrame1:AddMessage(format("%s HealthUpdated %i/%i", guid, health, healthMax))
	end

--]]

local MAJOR, MINOR = "LibQuickHealth-2.0", 2
local QuickHealth, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

QuickHealth.events = QuickHealth.events or LibStub("CallbackHandler-1.0"):New(QuickHealth)

local healthFromGUID = {}
local healthMaxFromGUID = {}
local healthQueueFromGUID = {}

QuickHealth.eventHandlers = QuickHealth.eventHandlers or {}
local eventHandlers = QuickHealth.eventHandlers
local eventFrame
local initialized

-- Public API functions.
--[[--------------------------------------------------------------------
UnitHealth("unitID")
Notes:
	Returns the health of the unit.
	Uses function notation instead of method notation to make it easy 
	to replace Blizz's UnitHealth.
Arguments: 
	string - UnitID of the unit to get health for.
Example:
	-- AddOns can use the following to optionally support LQH:
	local QuickHealth = LibStub and LibStub("LibQuickHealth-2.0", true)
	local UnitHealth = QuickHealth and QuickHealth.UnitHealth or UnitHealth
----------------------------------------------------------------------]]
function QuickHealth.UnitHealth(unitID)
	local guid = UnitGUID(unitID)
	return guid and healthFromGUID[guid] or UnitHealth(unitID)
end

--[[--------------------------------------------------------------------
:GetHealth("unitID" or "GUID")
Notes:
	Returns the health of the unit.
Arguments: 
	string - UnitID or GUID of the unit to get health for.
Example:
	local QuickHealth = LibStub("LibQuickHealth-2.0")
	local health = QuickHealth:GetHealth("player")
	assert( health == QuickHealth:GetHealth( UnitGUID("player") ) )
----------------------------------------------------------------------]]
function QuickHealth:GetHealth(unitID)
	local guid = unitID
	local health = healthFromGUID[guid]
	if not health then
		guid = UnitGUID(unitID)
		health = guid and healthFromGUID[guid]
	end
	return health or UnitHealth(unitID)
end

-- #NODOC
function QuickHealth:UnitHealthQueue(unitID)
	local guid = unitID
	local queue = healthQueueFromGUID[guid]
	if not queue then
		guid = UnitGUID(unitID)
		queue = guid and healthQueueFromGUID[guid]
	end
	if queue then
		return unpack(queue)
	end
end

-- Implementation.
if(QuickHealth.eventFrame) then
	eventFrame = QuickHealth.eventFrame
else
	eventFrame = CreateFrame("Frame")
	QuickHealth.eventFrame = eventFrame
	eventFrame:SetScript("OnEvent", function(self, event, ...) eventHandlers[event](...) end)
end

local function clear(array)
	for i=#array,1,-1 do
		array[i]=nil
	end
end

-- Map that will contain the units of every guid
local unitIDsFromGUID = {}
local RemapGUIDMap
do
	local updatedUnitIDs = {
		"player", "pet", 
		-- TODO "target", "focus", -- needs special event handling
	}
	for i=1,4 do
		table.insert(updatedUnitIDs, 'party'..i)
		table.insert(updatedUnitIDs, 'partypet'..i)
	end
	for i=1,40 do
		table.insert(updatedUnitIDs, 'raid'..i)
		table.insert(updatedUnitIDs, 'raidpet'..i)
	end

	-- This function will regenerate the map
	function RemapGUIDMap()
		for key in pairs(unitIDsFromGUID) do
			unitIDsFromGUID[key] = nil
		end
		for i,unitID in ipairs(updatedUnitIDs) do
			local guid = UnitGUID(unitID)
			if guid then
				if not unitIDsFromGUID[guid] then
					unitIDsFromGUID[guid] = {}
				end
				table.insert(unitIDsFromGUID[guid], unitID)
			end
		end
	end
end
-- TODO Remove
QuickHealth.guidmap = unitIDsFromGUID

eventHandlers.PARTY_MEMBERS_CHANGED = RemapGUIDMap
eventHandlers.RAID_ROSTER_UPDATE = RemapGUIDMap
eventHandlers.UNIT_PET = RemapGUIDMap -- TODO


local HealthUpdatedRegistered
local HealthUpdatedDebugRegistered
local UnitHealthUpdatedRegistered

local GetFrameNumber
do
	local counter = 0
	function GetFrameNumber() 
		return counter 
	end
	local function IncrementCounter()
		counter = counter+1 
	end

	-- When an addon starts listening to HealthUpdated we register our events and actually start doing something
	function QuickHealth.events:OnUsed(target, event)
		eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		eventFrame:RegisterEvent("UNIT_MAXHEALTH")
		eventFrame:RegisterEvent("UNIT_HEALTH")
		eventFrame:RegisterEvent("PLAYER_UNGHOST")

		if event == "HealthUpdated" then
			HealthUpdatedRegistered = true
		elseif event == "HealthUpdatedDebug" then
			-- Debug counter to know what events fired in the same frame.
			counter = 0
			eventFrame:SetScript("OnUpdate", IncrementCounter)
			HealthUpdatedDebugRegistered = true
		elseif event == "UnitHealthUpdated" then
			eventFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
			eventFrame:RegisterEvent("RAID_ROSTER_UPDATE")
			eventFrame:RegisterEvent("UNIT_PET")
			UnitHealthUpdatedRegistered = true
		end

		if not initialized then
			if UnitHealth('player') and eventHandlers.PLAYER_LOGIN then
				eventHandlers.PLAYER_LOGIN()
			else
				eventFrame:RegisterEvent("PLAYER_LOGIN")
			end
		end
	end

	-- Unregister the events and empty the healthtable
	function QuickHealth.events:OnUnused(target, event)
		if event == "HealthUpdated" then
			HealthUpdatedRegistered = nil
		elseif event == "HealthUpdatedDebug" then
			counter = nil
			eventFrame:SetScript("OnUpdate", nil)
			HealthUpdatedDebugRegistered = nil
		elseif event == "UnitHealthUpdated" then
			eventFrame:UnregisterEvent("PARTY_MEMBERS_CHANGED")
			eventFrame:UnregisterEvent("RAID_ROSTER_UPDATE")
			eventFrame:UnregisterEvent("UNIT_PET")
			UnitHealthUpdatedRegistered = nil
		end

		if not (HealthUpdatedRegistered or UnitHealthUpdatedRegistered or HealthUpdatedDebugRegistered) then
			eventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
			eventFrame:UnregisterEvent("UNIT_MAXHEALTH")
			eventFrame:UnregisterEvent("UNIT_HEALTH")
			eventFrame:UnregisterEvent("PLAYER_UNGHOST")
		end
	end
end

local MAX_QUEUE_SIZE = 5
local MAX_HP5_TICK = 20 -- 50 HP5

local function processUH(guid, unitHealth, unitHealthMax)
	if unitHealthMax == 100 then
		-- We might get here for unitID == target
		healthFromGUID[guid] = nil
		healthMaxFromGUID[guid] = nil
		healthQueueFromGUID[guid] = nil
		return
	end
	
	-- Get the current health from our table.
	local health = healthFromGUID[guid]
	if not health then
		-- Fill the health & healthMax tables and init an empty queue table.
		healthFromGUID[guid] = unitHealth
		healthMaxFromGUID[guid] = unitHealthMax
		healthQueueFromGUID[guid] = {}
		if HealthUpdatedRegistered then
			QuickHealth.events:Fire("HealthUpdated", guid, unitHealth, unitHealthMax)
		end
		if UnitHealthUpdatedRegistered then
			local unitID = unitIDsFromGUID[guid]
			if unitID then
				for i=1,#unitID do
					QuickHealth.events:Fire("UnitHealthUpdated", unitID[i], unitHealth, unitHealthMax)
				end
			end
		end
		return
	end

	local healthMax = healthMaxFromGUID[guid]
	local healthQueue = healthQueueFromGUID[guid]
	local fire
	if healthMax ~= unitHealthMax then
		-- TODO If healthMax changes, one of three things can happen:
		-- 1. health doesn't change (except if health>unitHealthMax ofc)
		-- 2. health changes by unitHealthMax-healthMax
		-- 3. health/healthMax doesn't change: h1/m1 = h2/m2, h2 = h1*m2/m1, h2-h1 = h1*(m2/m1-1)
		healthMax = unitHealthMax
		healthMaxFromGUID[guid] = unitHealthMax
		-- For now, just reset and hope it's correct. 
		health = unitHealth
		healthFromGUID[guid] = unitHealth
		fire = true
	end

	-- Ex unitHealth = 104, health = 100
	local discrepancy = unitHealth - health
	if discrepancy == 0 then
		clear(healthQueue)
	elseif #healthQueue==0 then
		health = unitHealth
		healthFromGUID[guid] = unitHealth
		fire = true
	elseif 0<discrepancy and discrepancy<=MAX_HP5_TICK then
		clear(healthQueue)
		health = unitHealth
		healthFromGUID[guid] = health
		fire = true
	else
		local noexplanation = true
		fire = true

		--local numdamage = #healthQueue--0
		local pendingDamage
		local healthChanged
		do
			-- Case 1: The last damage entries are pending.
			-- Ex. health = 1000-200-100+250 = 950, healthMax = 1000, unitHealth = 1000, 
			-- queue = (-200, -100, +250) = (-200, +200, -100) -> (-100)
			local h=health
			local pending = 0
			local overheal = 0
			for i=#healthQueue, 1, -1 do
				local amount = healthQueue[i]
				--if amount<0 then
					--numdamage = numdamage+1
					-- We assume that amount is pending, so it didn't count yet.
					-- Ex. i=2: h = 1000-(-100) = 1100
					h = h - amount
					pending = pending + amount
					-- Check for possible OHs.
					if h > healthMax then
						-- Ex. i=2: h=1000
						overheal = overheal + h-healthMax
						h = healthMax
					end
					local diff = unitHealth - h
					if 0<=diff and diff<=MAX_HP5_TICK then
						noexplanation = nil
						if diff>0 or overheal>0 then
							-- If we get here, our explanation is that an HP5 tick occured.
							healthChanged = true
						end
						pendingDamage = pending
						break
					end
				--end
			end
		end

		if noexplanation and --[[numdamage]] #healthQueue>1 then
			-- Case 2: A single damage entry is pending, but not the last one.
			for i=#healthQueue, 1, -1 do
				local amount = healthQueue[i]
				--if amount<0 then
					-- See what happens if amount is pending.
					local h = health - amount
					-- Check for OH.
					if h>healthMax then
						h = healthMax
					end
					local diff = unitHealth - h
					if 0<=diff and diff<=MAX_HP5_TICK then
						noexplanation = nil
						if diff>0 then
							-- If we get here, our explanation is that an HP5 tick occured.
							healthChanged = true
						end
						pendingDamage = amount
						break
					end
				--end
			end
		end

		if pendingDamage then
			-- Remove everything except the pending damage.
			healthQueue[1] = pendingDamage
			for j=#healthQueue, 2, -1 do
				healthQueue[j]=nil
			end
			if healthChanged then
				health = unitHealth + pendingDamage
				healthFromGUID[guid] = health
			end
		else
		--if noexplanation then
			-- Case 3: Nothing's pending, but the order of the combat log entries might have been wrong.
			-- If this results in more overheal than predicted, our prediction was wrong.
			-- We can't do anything about that. Just reset.
			-- Examples: 
			-- 19,179: (-a,+b,-c,+d) = (+b OH, -a, -c, +d) -> ()
			-- 19,1169: (-a,+b) = (+b OH,-a) -> ()
			-- 19,1426: (-a,-b,+c) = (-a,+c OH,-b) -> ()
			-- 19,1253: (-a pending, +b) = (+b OH, -a) -> ()
			-- 19,1942: (-a pending, +b,+c) = (+b,+c OH, -a) -> ()   [can't be (+b OH,-a,+c)]
			clear(healthQueue)
			health = unitHealth
			healthFromGUID[guid] = unitHealth
		end
	end

	if fire then
		if HealthUpdatedRegistered then
			QuickHealth.events:Fire("HealthUpdated", guid, health, healthMax)
		end
		if UnitHealthUpdatedRegistered then
			local unitID = unitIDsFromGUID[guid]
			if unitID then
				for i=1,#unitID do
					QuickHealth.events:Fire("UnitHealthUpdated", unitID[i], health, healthMax)
				end
			end
		end
	end
end

-- Amount is just the raw value from the combat log, including overheal.
local function processCLEU(guid, amount, amountMax)
	local health = healthFromGUID[guid]
	if not health then return end
	local healthQueue = healthQueueFromGUID[guid]
	local healthMax = healthMaxFromGUID[guid]

	if amountMax then
		healthMax = healthMax+amountMax
		healthMaxFromGUID[guid] = healthMax
	end

	local healthDiff = amount
	if health+amount>healthMax then
		healthDiff = healthMax-health
	end

	if healthDiff~=0 then
		if healthDiff<0 then
			-- We don't need to keep track of heals, because they cannot be pending anyways.
			tinsert(healthQueue, healthDiff)
			if #healthQueue>MAX_QUEUE_SIZE then
				local first = table.remove(healthQueue, 1)
				healthQueue[1] = healthQueue[1]+first
			end
		end
		health = health+healthDiff
		healthFromGUID[guid] = health
		if HealthUpdatedRegistered then
			QuickHealth.events:Fire("HealthUpdated", guid, health, healthMax)
		end
		if UnitHealthUpdatedRegistered then
			local unitID = unitIDsFromGUID[guid]
			if unitID then
				for i=1,#unitID do
					QuickHealth.events:Fire("UnitHealthUpdated", unitID[i], health, healthMax)
				end
			end
		end
	end
end

eventHandlers.CLEU = processCLEU
eventHandlers.UH = processUH


function eventHandlers.PLAYER_LOGIN()
	for _,unitID in pairs({"player","pet"}) do
		if UnitHealth(unitID) and UnitGUID(unitID) then
			initialized = true
			eventHandlers.UNIT_HEALTH(unitID)
		end
	end
	if initialized then
		RemapGUIDMap()
		eventHandlers.PLAYER_LOGIN = nil
	end
end

function eventHandlers.PLAYER_UNGHOST()
	eventHandlers.UNIT_HEALTH("player")
end

function eventHandlers.UNIT_HEALTH(unitID)
	if unitID == 'target' or unitID == 'focus' or unitID == 'mouseover' then 
		-- TODO
		if UnitHealthUpdatedRegistered then
			QuickHealth.events:Fire("UnitHealthUpdated", unitID, UnitHealth(unitID), UnitHealthMax(unitID))
		end
		return 
	end
	local unitHealth = UnitHealth(unitID)
	local unitHealthMax = UnitHealthMax(unitID)

	if HealthUpdatedDebugRegistered then
		QuickHealth.events:Fire("HealthUpdatedDebug", UnitGUID(unitID), 1, GetTime(), GetFrameNumber(), unitHealth, unitHealthMax)
	end

	processUH(UnitGUID(unitID), unitHealth, unitHealthMax)
end

function eventHandlers.UNIT_MAXHEALTH(unitID)
	if unitID == 'target' or unitID == 'focus' or unitID == 'mouseover' then 
		-- TODO
		if UnitHealthUpdatedRegistered then
			QuickHealth.events:Fire("UnitHealthUpdated", unitID, UnitHealth(unitID), UnitHealthMax(unitID))
		end
		return 
	end
	local unitHealth = UnitHealth(unitID)
	local unitHealthMax = UnitHealthMax(unitID)
	if HealthUpdatedDebugRegistered then
		QuickHealth.events:Fire("HealthUpdatedDebug", UnitGUID(unitID), 2, GetTime(), GetFrameNumber(), unitHealth, unitHealthMax)
	end

	processUH(UnitGUID(unitID), unitHealth, unitHealthMax)
end

-- For an event to be interesting, the unitFlags have to match *both* of the following bitsets. The reason we can't just combine
-- them into one bitset is that raid totems have COMBATLOG_OBJECT_AFFILIATION_RAID set, but not PLAYER or PET.
local flagRaid = bit.bor(COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_AFFILIATION_RAID)
local flagPlayerPets = bit.bor(COMBATLOG_OBJECT_TYPE_PLAYER, COMBATLOG_OBJECT_TYPE_PET)
local bitband = bit.band
function eventHandlers.COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	local amount
	local amountMax
	local guid
	
	if bitband(destFlags, flagRaid)>0 and bitband(destFlags, flagPlayerPets)>0 then
		-- We only care about damage or healing, check for either of those and react accordingly
		if(event == "SWING_DAMAGE") then --autoattack dmg
			amount = -(...) -- putting in braces will autoselect the first arg, no need to use select(1, ...)
		elseif(event == "SPELL_PERIODIC_DAMAGE" or event == "SPELL_DAMAGE"
			or event == "DAMAGE_SPLIT" or event == "DAMAGE_SHIELD") then -- all kinds of spelldamage
			amount = -select(4, ...)
		elseif(event == "SPELL_HEAL" or event == "SPELL_PERIODIC_HEAL") then --healing
			amount = select(4, ...)
		elseif(event == "ENVIRONMENTAL_DAMAGE") then --environmental damage
			amount = -select(2, ...)
--[[
		else
			local applied = event == "SPELL_AURA_APPLIED"
			if applied or event == "SPELL_AURA_REMOVED" then
				local spellID = (...)
				if spellID == 469 then
					-- TODO Commanding Shout
					local health = healthFromGUID[destGUID]
					if health then
						local healthMax = healthMaxFromGUID[destGUID]
						local p = health/healthMax
						if applied then
							amount = math.floor((healthMax+1080)*p + 0.5)
							amountMax = 1080
						else
							amount = math.floor((healthMax-1080)*p + 0.5)
							amountMax = -1080
						end
					end
				end
			end
--]]
		end
		guid = destGUID
	elseif bitband(sourceFlags, flagRaid)>0 and bitband(sourceFlags, flagPlayerPets)>0 then
		if event == "SPELL_CAST_SUCCESS" then
			local spellID = (...)
			if spellID==12975 then
				-- Last Stand. This should be pretty exact because healthMax doesn't change.
				local healthMax = healthMaxFromGUID[sourceGUID]
				if healthMax then
					amount = math.floor(0.3 * healthMax + 0.5)
					amountMax = amount
				end
--[[
			elseif spellID==2687 then
				-- TODO Bloodrage
				amount = -711
--]]
			end
		end
		guid = sourceGUID
	end

	if amount then
		if HealthUpdatedDebugRegistered then
			QuickHealth.events:Fire("HealthUpdatedDebug", guid, 3, GetTime(), GetFrameNumber(), amount, amountMax)
		end
		processCLEU(guid, amount, amountMax)
	end
end
